"""Parsers of StagYY output files.

Note:
    These functions are low level utilities. You should not use these unless
    you know what you are doing. To access StagYY output data, use an instance
    of :class:`~stagpy.stagyydata.StagyyData`.
"""
from functools import partial
from itertools import product
from operator import itemgetter
from xml.etree import ElementTree as xmlET
import re

import numpy as np
import pandas as pd
import h5py

from .error import ParsingError
from .phyvars import FIELD_FILES_H5, SFIELD_FILES_H5


def _tidy_names(names, nnames, extra_names=None):
    """Truncate or extend names so that its len is nnames.

    The list is modified, this function returns nothing.

    Args:
        names (list): list of names.
        nnames (int): desired number of names.
        extra_names (list of str): list of names to be used to extend the list
            if needed. If this list isn't provided, a range is used instead.
    """
    if len(names) < nnames and extra_names is not None:
        names.extend(extra_names)
    names.extend(range(nnames - len(names)))
    del names[nnames:]


def time_series(timefile, colnames):
    """Read temporal series text file.

    If :data:`colnames` is too long, it will be truncated. If it is too short,
    additional numeric column names from 0 to N-1 will be attributed to the N
    extra columns present in :data:`timefile`.

    Args:
        timefile (:class:`pathlib.Path`): path of the time.dat file.
        colnames (list of names): names of the variables expected in
            :data:`timefile` (may be modified).

    Returns:
        :class:`pandas.DataFrame`:
            Time series, with the variables in columns and the time steps in
            rows.
    """
    if not timefile.is_file():
        return None
    data = pd.read_csv(timefile, delim_whitespace=True, dtype=str,
                       header=None, skiprows=1, index_col=0,
                       engine='c', memory_map=True,
                       error_bad_lines=False, warn_bad_lines=False)
    data = data.apply(pd.to_numeric, raw=True, errors='coerce')

    # detect useless lines produced when run is restarted
    rows_to_del = []
    irow = len(data) - 1
    while irow > 0:
        iprev = irow - 1
        while iprev >= 0 and data.index[irow] <= data.index[iprev]:
            rows_to_del.append(iprev)
            iprev -= 1
        irow = iprev
    if rows_to_del:
        rows_to_keep = set(range(len(data))) - set(rows_to_del)
        data = data.take(list(rows_to_keep))

    ncols = data.shape[1]
    _tidy_names(colnames, ncols)
    data.columns = colnames

    return data


def time_series_h5(timefile, colnames):
    """Read temporal series HDF5 file.

    If :data:`colnames` is too long, it will be truncated. If it is too short,
    additional column names will be deduced from the content of the file.

    Args:
        timefile (:class:`pathlib.Path`): path of the TimeSeries.h5 file.
        colnames (list of names): names of the variables expected in
            :data:`timefile` (may be modified).

    Returns:
        :class:`pandas.DataFrame`:
            Time series, with the variables in columns and the time steps in
            rows.
    """
    if not timefile.is_file():
        return None
    with h5py.File(timefile, 'r') as h5f:
        dset = h5f['tseries']
        _, ncols = dset.shape
        ncols -= 1  # first is istep
        h5names = h5f['names'].asstr()[len(colnames) + 1:]
        _tidy_names(colnames, ncols, h5names)
        data = dset[()]
    pdf = pd.DataFrame(data[:, 1:],
                       index=np.int_(data[:, 0]), columns=colnames)
    # remove duplicated lines in case of restart
    return pdf.loc[~pdf.index.duplicated(keep='last')]


def _extract_rsnap_isteps(rproffile, data):
    """Extract istep, time and build separate rprof df."""
    step_regex = re.compile(r'^\*+step:\s*(\d+) ; time =\s*(\S+)')
    isteps = []  # list of (istep, time, df)
    line = ' '
    with rproffile.open() as stream:
        while line[0] != '*':
            line = stream.readline()
        match = step_regex.match(line)
        istep = int(match.group(1))
        time = float(match.group(2))
        nlines = 0
        iline = 0
        for line in stream:
            if line[0] == '*':
                isteps.append((istep, time, data.iloc[iline - nlines:iline]))
                match = step_regex.match(line)
                istep = int(match.group(1))
                time = float(match.group(2))
                nlines = 0
                # remove useless lines produced when run is restarted
                while isteps and istep <= isteps[-1][0]:
                    isteps.pop()
            else:
                nlines += 1
                iline += 1
        isteps.append((istep, time, data.iloc[iline - nlines:iline]))
    return isteps


def rprof(rproffile, colnames):
    """Extract radial profiles data.

    If :data:`colnames` is too long, it will be truncated. If it is too short,
    additional numeric column names from 0 to N-1 will be attributed to the N
    extra columns present in :data:`timefile`.

    Args:
        rproffile (:class:`pathlib.Path`): path of the rprof.dat file.
        colnames (list of names): names of the variables expected in
            :data:`rproffile`.

    Returns:
        tuple: (profs, times)
            :data:`profs` is a dict mapping istep to radial profiles
            :class:`pandas.DataFrame`.

            :data:`times` is the dimensionless time indexed by time steps.
    """
    if not rproffile.is_file():
        return {}, None
    data = pd.read_csv(rproffile, delim_whitespace=True, dtype=str,
                       header=None, comment='*', skiprows=1,
                       engine='c', memory_map=True,
                       error_bad_lines=False, warn_bad_lines=False)
    data = data.apply(pd.to_numeric, raw=True, errors='coerce')

    isteps = _extract_rsnap_isteps(rproffile, data)

    data = {}
    for istep, _, step_df in isteps:
        step_df.index = range(step_df.shape[0])  # check whether necessary
        step_cols = list(colnames)
        _tidy_names(step_cols, step_df.shape[1])
        step_df.columns = step_cols
        data[istep] = step_df

    df_times = pd.DataFrame(list(map(itemgetter(1), isteps)),
                            index=map(itemgetter(0), isteps))
    return data, df_times


def rprof_h5(rproffile, colnames):
    """Extract radial profiles data.

    If :data:`colnames` is too long, it will be truncated. If it is too short,
    additional column names will be deduced from the content of the file.

    Args:
        rproffile (:class:`pathlib.Path`): path of the rprof.dat file.
        colnames (list of names): names of the variables expected in
            :data:`rproffile`.

    Returns:
        tuple: (profs, times)
            :data:`profs` is a dict mapping istep to radial profiles
            :class:`pandas.DataFrame`.

            :data:`times` is the dimensionless time indexed by time steps.
    """
    if not rproffile.is_file():
        return {}, None
    isteps = []
    data = {}
    with h5py.File(rproffile, 'r') as h5f:
        dnames = sorted(dname for dname in h5f.keys()
                        if dname.startswith('rprof_'))
        h5names = h5f['names'].asstr()[len(colnames):]
        for dname in dnames:
            dset = h5f[dname]
            arr = dset[()]
            istep = dset.attrs['istep']
            step_cols = list(colnames)
            _tidy_names(step_cols, arr.shape[1], h5names)  # check shape
            data[istep] = pd.DataFrame(arr, columns=step_cols)
            isteps.append((istep, dset.attrs['time']))

    df_times = pd.DataFrame(list(map(itemgetter(1), isteps)),
                            index=map(itemgetter(0), isteps))
    return data, df_times


def _clean_names_refstate(names):
    """Uniformization of refstate profile names."""
    to_clean = {
        'Tref': 'T',
        'rhoref': 'rho',
        'tcond': 'Tcond',
    }
    return [to_clean.get(n, n) for n in names]


def refstate(reffile, ncols=8):
    """Extract reference state profiles.

    Args:
        reffile (:class:`pathlib.Path`): path of the refstate file.
        ncols (int): number of columns.

    Returns:
        tuple of list: (syst, adia)
            :data:`syst` is a list of list of :class:`pandas.DataFrame`
            containing the reference state profiles for each system and
            each phase in these systems.

            :data:`adia` is a list of :class:`pandas.DataFrame` containing
            the adiabatic reference state profiles for each system, the last
            item being the combined adiabat.
    """
    if not reffile.is_file():
        return None, None
    data = pd.read_csv(reffile, delim_whitespace=True, dtype=str,
                       header=None, names=range(ncols),
                       engine='c', memory_map=True,
                       error_bad_lines=False, warn_bad_lines=False)
    data = data.apply(pd.to_numeric, raw=True, errors='coerce')
    # drop lines corresponding to metadata
    data.dropna(subset=[0], inplace=True)
    isystem = -1
    systems = [[]]
    adiabats = []
    with reffile.open() as rsf:
        for line in rsf:
            line = line.lstrip()
            if line.startswith('SYSTEM'):
                isystem += 1
                if isystem > 0:
                    systems.append([])
            elif line.startswith('z'):
                systems[isystem].append(_clean_names_refstate(line.split()))
            elif line.startswith('ADIABAT') or line.startswith('COMBINED'):
                line = line.partition(':')[-1]
                adiabats.append(_clean_names_refstate(line.split()))
    nprofs = sum(map(len, systems)) + len(adiabats)
    nzprof = len(data) // nprofs
    iprof = 0
    syst = []
    adia = []
    for isys, layers in enumerate(systems):
        syst.append([])
        for layer in layers:
            ibgn = iprof * nzprof
            iend = ibgn + nzprof
            syst[isys].append(
                pd.DataFrame(data.iloc[ibgn:iend, :len(layer)].values,
                             columns=layer))
            iprof += 1
        if len(layers) > 1:
            ibgn = iprof * nzprof
            iend = ibgn + nzprof
            cols = adiabats.pop(0)
            adia.append(
                pd.DataFrame(data.iloc[ibgn:iend, :len(cols)].values,
                             columns=cols))
            iprof += 1
        else:
            adia.append(syst[isys][0])
    ibgn = iprof * nzprof
    iend = ibgn + nzprof
    cols = adiabats.pop(0)
    adia.append(pd.DataFrame(data.iloc[ibgn:iend, :len(cols)].values,
                             columns=cols))
    return syst, adia


def _readbin(fid, fmt='i', nwords=1, file64=False, unpack=True):
    """Read n words of 4 or 8 bytes with fmt format.

    fmt: 'i' or 'f' or 'b' (integer or float or bytes)
    4 or 8 bytes: depends on header

    Return an array of elements if more than one element.

    Default: read 1 word formatted as an integer.
    """
    if fmt in 'if':
        fmt += '8' if file64 else '4'
    elts = np.fromfile(fid, fmt, nwords)
    if unpack and len(elts) == 1:
        elts = elts[0]
    return elts


def fields(fieldfile, only_header=False, only_istep=False):
    """Extract fields data.

    Args:
        fieldfile (:class:`pathlib.Path`): path of the binary field file.
        only_header (bool): when True (and :data:`only_istep` is False), only
            :data:`header` is returned.
        only_istep (bool): when True, only :data:`istep` is returned.

    Returns:
        depends on flags.: :obj:`int`: istep
            If :data:`only_istep` is True, this function returns the time step
            at which the binary file was written.
        :obj:`dict`: header
            Else, if :data:`only_header` is True, this function returns a dict
            containing the header informations of the binary file.
        :class:`numpy.array`: fields
            Else, this function returns the tuple :data:`(header, fields)`.
            :data:`fields` is an array of scalar fields indexed by variable,
            x-direction, y-direction, z-direction, block.
    """
    # something to skip header?
    if not fieldfile.is_file():
        return None
    header = {}
    with fieldfile.open('rb') as fid:
        readbin = partial(_readbin, fid)
        magic = readbin()
        if magic > 8000:  # 64 bits
            magic -= 8000
            readbin()  # need to read 4 more bytes
            readbin = partial(readbin, file64=True)

        # check nb components
        nval = 1
        sfield = False
        if magic > 400:
            nval = 4
        elif magic > 300:
            nval = 3
        elif magic > 100:
            sfield = True

        magic %= 100
        if magic < 9:
            raise ParsingError(fieldfile, 'magic < 9 not supported')

        # extra ghost point in horizontal direction
        header['xyp'] = int(nval == 4)  # magic >= 9

        # total number of values in relevant space basis
        # (e1, e2, e3) = (theta, phi, radius) in spherical geometry
        #              = (x, y, z)            in cartesian geometry
        header['nts'] = readbin(nwords=3)

        # number of blocks, 2 for yinyang or cubed sphere
        header['ntb'] = readbin()  # magic >= 7

        # aspect ratio
        header['aspect'] = readbin('f', 2)

        # number of parallel subdomains
        header['ncs'] = readbin(nwords=3)  # (e1, e2, e3) space
        header['ncb'] = readbin()  # magic >= 8, blocks

        # r - coordinates
        # rgeom[0:self.nrtot+1, 0] are edge radial position
        # rgeom[0:self.nrtot, 1] are cell-center radial position
        header['rgeom'] = readbin('f', header['nts'][2] * 2 + 1)  # magic >= 2
        header['rgeom'] = np.resize(header['rgeom'], (header['nts'][2] + 1, 2))

        header['rcmb'] = readbin('f')  # magic >= 7

        header['ti_step'] = readbin()  # magic >= 3
        if only_istep:
            return header['ti_step']
        header['ti_ad'] = readbin('f')  # magic >= 3
        header['erupta_total'] = readbin('f')  # magic >= 5
        header['bot_temp'] = readbin('f')  # magic >= 6
        header['core_temp'] = readbin('f') if magic >= 10 else 1

        # magic >= 4
        header['e1_coord'] = readbin('f', header['nts'][0])
        header['e2_coord'] = readbin('f', header['nts'][1])
        header['e3_coord'] = readbin('f', header['nts'][2])

        if only_header:
            return header

        # READ FIELDS
        # number of points in (e1, e2, e3) directions PER CPU
        npc = header['nts'] // header['ncs']
        # number of blocks per cpu
        nbk = header['ntb'] // header['ncb']
        # number of values per 'read' block
        npi = (npc[0] + header['xyp']) * (npc[1] + header['xyp']) * npc[2] * \
            nbk * nval

        header['scalefac'] = readbin('f') if nval > 1 else 1

        flds = np.zeros((nval,
                         header['nts'][0] + header['xyp'],
                         header['nts'][1] + header['xyp'],
                         header['nts'][2],
                         header['ntb']))

        # loop over parallel subdomains
        for icpu in product(range(header['ncb']),
                            range(header['ncs'][2]),
                            range(header['ncs'][1]),
                            range(header['ncs'][0])):
            # read the data for one CPU
            data_cpu = readbin('f', npi) * header['scalefac']

            # icpu is (icpu block, icpu z, icpu y, icpu x)
            # data from file is transposed to obtained a field
            # array indexed with (x, y, z, block), as in StagYY
            flds[:,
                 icpu[3] * npc[0]:(icpu[3] + 1) * npc[0] + header['xyp'],  # x
                 icpu[2] * npc[1]:(icpu[2] + 1) * npc[1] + header['xyp'],  # y
                 icpu[1] * npc[2]:(icpu[1] + 1) * npc[2],  # z
                 icpu[0] * nbk:(icpu[0] + 1) * nbk  # block
                 ] = np.transpose(data_cpu.reshape(
                     (nbk, npc[2], npc[1] + header['xyp'],
                      npc[0] + header['xyp'], nval)))
        if sfield:
            # for surface fields, variables are written along z direction
            flds = np.swapaxes(flds, 0, 3)
    return header, flds


def tracers(tracersfile):
    """Extract tracers data.

    Args:
        tracersfile (:class:`pathlib.Path`): path of the binary tracers file.

    Returns:
        dict of list of numpy.array:
            Tracers data organized by attribute and block.
    """
    if not tracersfile.is_file():
        return None
    tra = {}
    with tracersfile.open('rb') as fid:
        readbin = partial(_readbin, fid)
        magic = readbin()
        if magic > 8000:  # 64 bits
            magic -= 8000
            readbin()
            readbin = partial(readbin, file64=True)
        if magic < 100:
            raise ParsingError(tracersfile,
                               'magic > 100 expected to get tracervar info')
        nblk = magic % 100
        readbin('f', 2)  # aspect ratio
        readbin()  # istep
        readbin('f')  # time
        ninfo = readbin()
        ntra = readbin(nwords=nblk, unpack=False)
        readbin('f')  # tracer ideal mass
        curv = readbin()
        if curv:
            readbin('f')  # r_cmb
        infos = []  # list of info names
        for _ in range(ninfo):
            infos.append(b''.join(readbin('b', 16)).strip().decode())
            tra[infos[-1]] = []
        if magic > 200:
            ntrace_elt = readbin()
            if ntrace_elt > 0:
                readbin('f', ntrace_elt)  # outgassed
        for ntrab in ntra:  # blocks
            data = readbin('f', ntrab * ninfo)
            for idx, info in enumerate(infos):
                tra[info].append(data[idx::ninfo])
    return tra


def _read_group_h5(filename, groupname):
    """Return group content.

    Args:
        filename (:class:`pathlib.Path`): path of hdf5 file.
        groupname (str): name of group to read.
    Returns:
        :class:`numpy.array`: content of group.
    """
    with h5py.File(filename, 'r') as h5f:
        data = h5f[groupname][()]
    return data  # need to be reshaped


def _make_3d(field, twod):
    """Add a dimension to field if necessary.

    Args:
        field (numpy.array): the field that need to be 3d.
        twod (str): 'XZ', 'YZ' or None depending on what is relevant.
    Returns:
        numpy.array: reshaped field.
    """
    shp = list(field.shape)
    if twod and 'X' in twod:
        shp.insert(1, 1)
    elif twod:
        shp.insert(0, 1)
    return field.reshape(shp)


def _ncores(meshes, twod):
    """Compute number of nodes in each direction."""
    nnpb = len(meshes)  # number of nodes per block
    nns = [1, 1, 1]  # number of nodes in x, y, z directions
    if twod is None or 'X' in twod:
        while (nnpb > 1 and
               meshes[nns[0]]['X'][0, 0, 0] ==
               meshes[nns[0] - 1]['X'][-1, 0, 0]):
            nns[0] += 1
            nnpb -= 1
    cpu = lambda icy: icy * nns[0]
    if twod is None or 'Y' in twod:
        while (nnpb > 1 and
               meshes[cpu(nns[1])]['Y'][0, 0, 0] ==
               meshes[cpu(nns[1] - 1)]['Y'][0, -1, 0]):
            nns[1] += 1
            nnpb -= nns[0]
    cpu = lambda icz: icz * nns[0] * nns[1]
    while (nnpb > 1 and
           meshes[cpu(nns[2])]['Z'][0, 0, 0] ==
           meshes[cpu(nns[2] - 1)]['Z'][0, 0, -1]):
        nns[2] += 1
        nnpb -= nns[0] * nns[1]
    return np.array(nns)


def _conglomerate_meshes(meshin, header):
    """Conglomerate meshes from several cores into one."""
    meshout = {}
    npc = header['nts'] // header['ncs']
    shp = [val + 1 if val != 1 else 1 for val in header['nts']]
    x_p = int(shp[0] != 1)
    y_p = int(shp[1] != 1)
    for coord in meshin[0]:
        meshout[coord] = np.zeros(shp)
    for icore in range(np.prod(header['ncs'])):
        ifs = [icore // np.prod(header['ncs'][:i]) % header['ncs'][i] * npc[i]
               for i in range(3)]
        for coord, mesh in meshin[icore].items():
            meshout[coord][ifs[0]:ifs[0] + npc[0] + x_p,
                           ifs[1]:ifs[1] + npc[1] + y_p,
                           ifs[2]:ifs[2] + npc[2] + 1] = mesh
    return meshout


def _read_coord_h5(files, shapes, header, twod):
    """Read all coord hdf5 files of a snapshot.

    Args:
        files (list of pathlib.Path): list of NodeCoordinates files of
            a snapshot.
        shapes (list of (int,int)): shape of mesh grids.
        header (dict): geometry info.
        twod (str): 'XZ', 'YZ' or None depending on what is relevant.
    """
    meshes = []
    for h5file, shape in zip(files, shapes):
        meshes.append({})
        with h5py.File(h5file, 'r') as h5f:
            for coord, mesh in h5f.items():
                # for some reason, the array is transposed!
                meshes[-1][coord] = mesh[()].reshape(shape).T
                meshes[-1][coord] = _make_3d(meshes[-1][coord], twod)

    header['ncs'] = _ncores(meshes, twod)
    header['nts'] = list((meshes[0]['X'].shape[i] - 1) * header['ncs'][i]
                         for i in range(3))
    header['nts'] = np.array([max(1, val) for val in header['nts']])
    meshes = _conglomerate_meshes(meshes, header)
    if np.any(meshes['Z'][:, :, 0] != 0):
        # spherical
        if twod is not None:  # annulus geometry...
            header['x_mesh'] = np.copy(meshes['Y'])
            header['y_mesh'] = np.copy(meshes['Z'])
            header['z_mesh'] = np.copy(meshes['X'])
        else:  # YinYang, here only yin
            header['x_mesh'] = np.copy(meshes['X'])
            header['y_mesh'] = np.copy(meshes['Y'])
            header['z_mesh'] = np.copy(meshes['Z'])
        header['r_mesh'] = np.sqrt(header['x_mesh']**2 + header['y_mesh']**2 +
                                   header['z_mesh']**2)
        header['t_mesh'] = np.arccos(header['z_mesh'] / header['r_mesh'])
        header['p_mesh'] = np.roll(
            np.arctan2(header['y_mesh'], -header['x_mesh']) + np.pi, -1, 1)
        header['e1_coord'] = header['t_mesh'][:, 0, 0]
        header['e2_coord'] = header['p_mesh'][0, :, 0]
        header['e3_coord'] = header['r_mesh'][0, 0, :]
    else:
        header['e1_coord'] = meshes['X'][:, 0, 0]
        header['e2_coord'] = meshes['Y'][0, :, 0]
        header['e3_coord'] = meshes['Z'][0, 0, :]
    header['aspect'] = (header['e1_coord'][-1] - header['e2_coord'][0],
                        header['e1_coord'][-1] - header['e2_coord'][0])
    header['rcmb'] = header['e3_coord'][0]
    if header['rcmb'] == 0:
        header['rcmb'] = -1
    else:
        header['e3_coord'] = header['e3_coord'] - header['rcmb']
    if twod is None or 'X' in twod:
        header['e1_coord'] = header['e1_coord'][:-1]
    if twod is None or 'Y' in twod:
        header['e2_coord'] = header['e2_coord'][:-1]
    header['e3_coord'] = header['e3_coord'][:-1]


def _get_dim(data_item):
    """Extract shape of data item."""
    return tuple(map(int, data_item.get('Dimensions').split()))


def _get_field(xdmf_file, data_item):
    """Extract field from data item."""
    shp = _get_dim(data_item)
    h5file, group = data_item.text.strip().split(':/', 1)
    # Field on yin is named <var>_XXXXX_YYYYY, on yang is <var>2XXXXX_YYYYY.
    numeral_part = group[-11:]
    icore = int(numeral_part.split('_')[-2]) - 1
    fld = None
    try:
        fld = _read_group_h5(xdmf_file.parent / h5file, group).reshape(shp)
    except KeyError:
        # test previous/following snapshot files as their numbers can get
        # slightly out of sync between cores
        h5file_parts = h5file.split('_')
        fnum = int(h5file_parts[-2])
        if fnum > 0:
            h5file_parts[-2] = f'{fnum - 1:05d}'
            h5f = xdmf_file.parent / '_'.join(h5file_parts)
            try:
                fld = _read_group_h5(h5f, group).reshape(shp)
            except (OSError, KeyError):
                pass
        if fld is None:
            h5file_parts[-2] = f'{fnum + 1:05d}'
            h5f = xdmf_file.parent / '_'.join(h5file_parts)
            try:
                fld = _read_group_h5(h5f, group).reshape(shp)
            except (OSError, KeyError):
                pass
        if fld is None:
            raise
    return icore, fld


def _maybe_get(elt, item, info, conversion=None):
    """Extract and convert info if item is present."""
    maybe_item = elt.find(item)
    if maybe_item is not None:
        maybe_item = maybe_item.get(info)
        if conversion is not None:
            maybe_item = conversion(maybe_item)
    return maybe_item


def read_geom_h5(xdmf_file, snapshot):
    """Extract geometry information from hdf5 files.

    Args:
        xdmf_file (:class:`pathlib.Path`): path of the xdmf file.
        snapshot (int): snapshot number.
    Returns:
        (dict, root): geometry information and root of xdmf document.
    """
    header = {}
    xdmf_root = xmlET.parse(str(xdmf_file)).getroot()
    if snapshot is None:
        return None, xdmf_root

    # Domain, Temporal Collection, Snapshot
    # should check that this is indeed the required snapshot
    elt_snap = xdmf_root[0][0][snapshot]
    header['ti_ad'] = float(elt_snap.find('Time').get('Value'))
    header['mo_lambda'] = _maybe_get(elt_snap, 'mo_lambda', 'Value', float)
    header['mo_thick_sol'] = _maybe_get(elt_snap, 'mo_thick_sol', 'Value',
                                        float)
    header['ntb'] = 1
    coord_h5 = []  # all the coordinate files
    coord_shape = []  # shape of meshes
    twod = None
    for elt_subdomain in elt_snap.findall('Grid'):
        if elt_subdomain.get('Name').startswith('meshYang'):
            header['ntb'] = 2
            break  # iterate only through meshYin
        elt_geom = elt_subdomain.find('Geometry')
        if elt_geom.get('Type') == 'X_Y' and twod is None:
            twod = ''
            for data_item in elt_geom.findall('DataItem'):
                coord = data_item.text.strip()[-1]
                if coord in 'XYZ':
                    twod += coord
        data_item = elt_geom.find('DataItem')
        coord_shape.append(_get_dim(data_item))
        coord_h5.append(
            xdmf_file.parent / data_item.text.strip().split(':/', 1)[0])
    _read_coord_h5(coord_h5, coord_shape, header, twod)
    return header, xdmf_root


def _to_spherical(flds, header):
    """Convert vector field to spherical."""
    cth = np.cos(header['t_mesh'][:, :, :-1])
    sth = np.sin(header['t_mesh'][:, :, :-1])
    cph = np.cos(header['p_mesh'][:, :, :-1])
    sph = np.sin(header['p_mesh'][:, :, :-1])
    fout = np.copy(flds)
    fout[0] = cth * cph * flds[0] + cth * sph * flds[1] - sth * flds[2]
    fout[1] = sph * flds[0] - cph * flds[1]  # need to take the opposite here
    fout[2] = sth * cph * flds[0] + sth * sph * flds[1] + cth * flds[2]
    return fout


def _flds_shape(fieldname, header):
    """Compute shape of flds variable."""
    shp = list(header['nts'])
    shp.append(header['ntb'])
    if len(FIELD_FILES_H5.get(fieldname, [])) == 3:
        shp.insert(0, 3)
        # extra points
        header['xp'] = int(header['nts'][0] != 1)
        shp[1] += header['xp']
        header['yp'] = int(header['nts'][1] != 1)
        shp[2] += header['yp']
        header['zp'] = 1
        header['xyp'] = 1
    else:
        shp.insert(0, 1)
        header['xp'] = 0
        header['yp'] = 0
        header['zp'] = 0
        header['xyp'] = 0
    return shp


def _post_read_flds(flds, header):
    """Process flds to handle sphericity."""
    if flds.shape[0] >= 3 and header['rcmb'] > 0:
        # spherical vector
        header['p_mesh'] = np.roll(
            np.arctan2(header['y_mesh'], header['x_mesh']), -1, 1)
        for ibk in range(header['ntb']):
            flds[..., ibk] = _to_spherical(flds[..., ibk], header)
        header['p_mesh'] = np.roll(
            np.arctan2(header['y_mesh'], -header['x_mesh']) + np.pi, -1, 1)
    return flds


def read_field_h5(xdmf_file, fieldname, snapshot, header=None):
    """Extract field data from hdf5 files.

    Args:
        xdmf_file (:class:`pathlib.Path`): path of the xdmf file.
        fieldname (str): name of field to extract.
        snapshot (int): snapshot number.
        header (dict): geometry information.
    Returns:
        (dict, numpy.array): geometry information and field data. None
            is returned if data is unavailable.
    """
    if header is None:
        header, xdmf_root = read_geom_h5(xdmf_file, snapshot)
    else:
        xdmf_root = xmlET.parse(str(xdmf_file)).getroot()

    npc = header['nts'] // header['ncs']  # number of grid point per node
    flds = np.zeros(_flds_shape(fieldname, header))
    data_found = False

    for elt_subdomain in xdmf_root[0][0][snapshot].findall('Grid'):
        ibk = int(elt_subdomain.get('Name').startswith('meshYang'))
        for data_attr in elt_subdomain.findall('Attribute'):
            if data_attr.get('Name') != fieldname:
                continue
            icore, fld = _get_field(xdmf_file, data_attr.find('DataItem'))
            # for some reason, the field is transposed
            fld = fld.T
            shp = fld.shape
            if shp[-1] == 1 and header['nts'][0] == 1:  # YZ
                fld = fld.reshape((shp[0], 1, shp[1], shp[2]))
                if header['rcmb'] < 0:
                    fld = fld[(2, 0, 1), ...]
            elif shp[-1] == 1:  # XZ
                fld = fld.reshape((shp[0], shp[1], 1, shp[2]))
                if header['rcmb'] < 0:
                    fld = fld[(1, 2, 0), ...]
            elif fieldname in SFIELD_FILES_H5:
                fld = fld.reshape((1, npc[0], npc[1], 1))
            elif header['nts'][1] == 1:  # cart XZ
                fld = fld.reshape((1, shp[0], 1, shp[1]))
            ifs = [icore // np.prod(header['ncs'][:i]) % header['ncs'][i] *
                   npc[i] for i in range(3)]
            if fieldname in SFIELD_FILES_H5:
                ifs[2] = 0
                npc[2] = 1
            if header['zp']:  # remove top row
                fld = fld[:, :, :, :-1]
            flds[:,
                 ifs[0]:ifs[0] + npc[0] + header['xp'],
                 ifs[1]:ifs[1] + npc[1] + header['yp'],
                 ifs[2]:ifs[2] + npc[2],
                 ibk] = fld
            data_found = True

    if flds.shape[0] == 3 and flds.shape[-1] == 2:  # YinYang vector
        # Yang grid is rotated compared to Yin grid
        flds[0, ..., 1] = -flds[0, ..., 1]
        vt = flds[1, ..., 1].copy()
        flds[1, ..., 1] = flds[2, ..., 1]
        flds[2, ..., 1] = vt
    flds = _post_read_flds(flds, header)

    if fieldname in SFIELD_FILES_H5:
        # remove z component
        flds = flds[..., 0, :]
    return (header, flds) if data_found else None


def read_tracers_h5(xdmf_file, infoname, snapshot, position):
    """Extract tracers data from hdf5 files.

    Args:
        xdmf_file (:class:`pathlib.Path`): path of the xdmf file.
        infoname (str): name of information to extract.
        snapshot (int): snapshot number.
        position (bool): whether to extract position of tracers.
    Returns:
        dict of list of numpy.array:
            Tracers data organized by attribute and block.
    """
    xdmf_root = xmlET.parse(str(xdmf_file)).getroot()
    tra = {}
    tra[infoname] = [{}, {}]  # two blocks, ordered by cores
    if position:
        for axis in 'xyz':
            tra[axis] = [{}, {}]
    for elt_subdomain in xdmf_root[0][0][snapshot].findall('Grid'):
        ibk = int(elt_subdomain.get('Name').startswith('meshYang'))
        if position:
            for data_attr in elt_subdomain.findall('Geometry'):
                for data_item, axis in zip(data_attr.findall('DataItem'),
                                           'xyz'):
                    icore, data = _get_field(xdmf_file, data_item)
                    tra[axis][ibk][icore] = data
        for data_attr in elt_subdomain.findall('Attribute'):
            if data_attr.get('Name') != infoname:
                continue
            icore, data = _get_field(xdmf_file, data_attr.find('DataItem'))
            tra[infoname][ibk][icore] = data
    for info in tra:
        tra[info] = [trab for trab in tra[info] if trab]  # remove empty blocks
        for iblk, trab in enumerate(tra[info]):
            tra[info][iblk] = np.concatenate([trab[icore]
                                              for icore in range(len(trab))])
    return tra


def read_time_h5(h5folder):
    """Iterate through (isnap, istep) recorded in h5folder/'time_botT.h5'.

    Args:
        h5folder (:class:`pathlib.Path`): directory of HDF5 output files.
    Yields:
        tuple of int: (isnap, istep).
    """
    with h5py.File(h5folder / 'time_botT.h5', 'r') as h5f:
        for name, dset in h5f.items():
            isnap = int(name[-5:])
            if len(dset) == 3:
                istep = int(dset[2])
            else:
                istep = int(dset[0])
            yield isnap, istep



from pkg_resources import safe_version
import stagpy


def test_version_format():
    assert stagpy.__version__ == safe_version(stagpy.__version__)


def test_bogus_mplstyle(capsys):
    stagpy.conf.plot.mplstyle = 'stagpy-bogus'
    stagpy.load_mplstyle()
    output = capsys.readouterr()
    assert output.err == 'Cannot import style stagpy-bogus.\n'
    del stagpy.conf.plot.mplstyle



"""Exceptions raised by StagPy."""


class StagpyError(Exception):
    """Base class for exceptions raised by StagPy.

    Note:
        All exceptions derive from this class. To catch any error that might be
        raised by StagPy due to invalid requests/missing data, you only need to
        catch this exception.
    """

    pass


class NoSnapshotError(StagpyError):
    """Raised when no snapshot can be found.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which no snapshot were found.

    Attributes:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which no snapshot were found.
    """

    def __init__(self, sdat):
        self.sdat = sdat
        super().__init__(f'no snapshot found for {sdat}')


class NoParFileError(StagpyError):
    """Raised when no par file can be found.

    Args:
        parfile (pathlike): the expected path of
            the par file.

    Attributes:
        parfile (pathlike): the expected path of the par file.
    """

    def __init__(self, parfile):
        self.parfile = parfile
        super().__init__(f'{parfile} file not found')


class NotAvailableError(StagpyError):
    """Raised when a feature is not available yet."""

    pass


class ParsingError(StagpyError):
    """Raised when a parsing error occurs.

    Args:
        faulty_file (pathlike): path of the file where a parsing problem
            was encountered.
        msg (str): error message.

    Attributes:
        file (pathlike): path of the file where a parsing problem was
            encountered.
        msg (str): error message.
    """

    def __init__(self, faulty_file, msg):
        self.file = faulty_file
        self.msg = msg
        super().__init__(faulty_file, msg)


class InvalidTimestepError(StagpyError, KeyError):
    """Raised when invalid time step is requested.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which the request was made.
        istep (int): the invalid time step index.
        msg (str): the error message.

    Attributes:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which the request was made.
        istep (int): the invalid time step index.
        msg (str): the error message.
    """

    def __init__(self, sdat, istep, msg):
        self.sdat = sdat
        self.istep = istep
        self.msg = msg
        super().__init__(sdat, istep, msg)


class InvalidSnapshotError(StagpyError, KeyError):
    """Raised when invalid snapshot is requested.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which the request was made.
        isnap (int): the invalid snapshot index.
        msg (str): the error message.

    Attributes:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which the request was made.
        isnap (int): the invalid snapshot index.
        msg (str): the error message.
    """

    def __init__(self, sdat, isnap, msg):
        self.sdat = sdat
        self.isnap = isnap
        self.msg = msg
        super().__init__(sdat, isnap, msg)


class InvalidTimeFractionError(StagpyError):
    """Raised when invalid fraction of series is requested.

    Args:
        fraction (float): the invalid fraction.

    Attributes:
        fraction (float): the invalid fraction.
    """

    def __init__(self, fraction):
        self.fraction = fraction
        super().__init__(f'Fraction should be in (0,1] (received {fraction})')


class InvalidNfieldsError(StagpyError):
    """Raised when invalid nfields_max is requested.

    Args:
        nfields (int): the invalid number of fields.

    Attributes:
        nfields (int): the invalid number of field.
    """

    def __init__(self, nfields):
        self.nfields = nfields
        super().__init__(f'nfields_max should be >5 (received {nfields})')


class InvalidZoomError(StagpyError):
    """Raised when invalid zoom is requested.

    Args:
        zoom (int): the invalid zoom level.

    Attributes:
        zoom (int): the invalid zoom level.
    """

    def __init__(self, zoom):
        self.zoom = zoom
        super().__init__(f'Zoom angle should be in [0,360] (received {zoom})')


class StagnantLidError(StagpyError):
    """Raised when unexpected stagnant lid regime is found.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which a stagnant lid regime was found.

    Attributes:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance for which a stagnant lid regime was found.
    """

    def __init__(self, sdat):
        self.sdat = sdat
        super().__init__(f'Stagnant lid regime for {sdat}')


class MissingDataError(StagpyError, KeyError):
    """Raised when requested data is not present in output."""

    pass


class UnknownVarError(StagpyError, KeyError):
    """Raised when invalid var is requested.

    Args:
        varname (str): the invalid var name.

    Attributes:
        varname (str): the invalid var name.
    """

    def __init__(self, varname):
        self.varname = varname
        super().__init__(varname)


class UnknownFiltersError(StagpyError):
    """Raised when invalid step filter is requested.

    Args:
        filters (list): the invalid filter names.

    Attributes:
        filters (list): the invalid filter names.
    """

    def __init__(self, filters):
        self.filters = filters
        super().__init__(', '.join(repr(f) for f in filters))


class UnknownFieldVarError(UnknownVarError):
    """Raised when invalid field var is requested.

    Derived from :class:`~stagpy.error.UnknownVarError`.
    """

    pass


class UnknownRprofVarError(UnknownVarError):
    """Raised when invalid rprof var is requested.

    Derived from :class:`~stagpy.error.UnknownVarError`.
    """

    pass


class UnknownTimeVarError(UnknownVarError):
    """Raised when invalid time var is requested.

    Derived from :class:`~stagpy.error.UnknownVarError`.
    """

    pass



"""Plot scalar and vector fields."""

from itertools import chain

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as mpat
from mpl_toolkits.axes_grid1 import make_axes_locatable

from . import conf, phyvars, _helpers
from .error import NotAvailableError
from .stagyydata import StagyyData


# The location is off for vertical velocities: they have an extra
# point in (x,y) instead of z in the output


def _threed_extract(step, var, walls=False):
    """Return suitable slices and coords for 3D fields."""
    is_vector = not valid_field_var(var)
    hwalls = is_vector or walls
    i_x = conf.field.ix
    i_y = conf.field.iy
    i_z = conf.field.iz
    if i_x is not None or i_y is not None:
        i_z = None
    if i_x is not None or i_z is not None:
        i_y = None
    if i_x is None and i_y is None and i_z is None:
        i_x = 0
    if i_x is not None:
        xcoord = step.geom.y_walls if hwalls else step.geom.y_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        i_y = i_z = slice(None)
        varx, vary = var + '2', var + '3'
    elif i_y is not None:
        xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        i_x = i_z = slice(None)
        varx, vary = var + '1', var + '3'
    else:
        xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
        ycoord = step.geom.y_walls if hwalls else step.geom.y_centers
        i_x = i_y = slice(None)
        varx, vary = var + '1', var + '2'
    if is_vector:
        data = (step.fields[varx][i_x, i_y, i_z, 0],
                step.fields[vary][i_x, i_y, i_z, 0])
    else:
        data = step.fields[var][i_x, i_y, i_z, 0]
    return (xcoord, ycoord), data


def valid_field_var(var):
    """Whether a field variable is defined.

    This function checks if a definition of the variable exists in
    :data:`~stagpy.phyvars.FIELD` or :data:`~stagpy.phyvars.FIELD_EXTRA`.

    Args:
        var (str): the variable name to be checked.
    Returns:
        bool: True is the var is defined, False otherwise.
    """
    return var in phyvars.FIELD or var in phyvars.FIELD_EXTRA


def get_meshes_fld(step, var, walls=False):
    """Return scalar field along with coordinates meshes.

    Only works properly in 2D geometry and 3D cartesian.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
        var (str): scalar field name.
        walls (bool): consider the walls as the relevant mesh.
    Returns:
        tuple of :class:`numpy.array`: xmesh, ymesh, fld
            2D arrays containing respectively the x position, y position, and
            the value of the requested field.
    """
    fld = step.fields[var]
    hwalls = (walls or fld.shape[0] != step.geom.nxtot or
              fld.shape[1] != step.geom.nytot)
    if step.geom.threed and step.geom.cartesian:
        (xcoord, ycoord), fld = _threed_extract(step, var, walls)
    elif step.geom.twod_xz:
        xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        fld = fld[:, 0, :, 0]
    else:  # twod_yz
        xcoord = step.geom.y_walls if hwalls else step.geom.y_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        if step.geom.curvilinear:
            pmesh, rmesh = np.meshgrid(xcoord, ycoord, indexing='ij')
            xmesh, ymesh = rmesh * np.cos(pmesh), rmesh * np.sin(pmesh)
        fld = fld[0, :, :, 0]
    if step.geom.cartesian:
        xmesh, ymesh = np.meshgrid(xcoord, ycoord, indexing='ij')
    return xmesh, ymesh, fld


def get_meshes_vec(step, var):
    """Return vector field components along with coordinates meshes.

    Only works properly in 2D geometry and 3D cartesian.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
        var (str): vector field name.
    Returns:
        tuple of :class:`numpy.array`: xmesh, ymesh, fldx, fldy
            2D arrays containing respectively the x position, y position, x
            component and y component of the requested vector field.
    """
    if step.geom.threed and step.geom.cartesian:
        (xcoord, ycoord), (vec1, vec2) = _threed_extract(step, var)
    elif step.geom.twod_xz:
        xcoord, ycoord = step.geom.x_walls, step.geom.z_centers
        vec1 = step.fields[var + '1'][:, 0, :, 0]
        vec2 = step.fields[var + '3'][:, 0, :, 0]
    elif step.geom.cartesian and step.geom.twod_yz:
        xcoord, ycoord = step.geom.y_walls, step.geom.z_centers
        vec1 = step.fields[var + '2'][0, :, :, 0]
        vec2 = step.fields[var + '3'][0, :, :, 0]
    else:  # spherical yz
        pcoord = step.geom.p_walls
        pmesh = np.outer(pcoord, np.ones(step.geom.nrtot))
        vec_phi = step.fields[var + '2'][0, :, :, 0]
        vec_r = step.fields[var + '3'][0, :, :, 0]
        vec1 = vec_r * np.cos(pmesh) - vec_phi * np.sin(pmesh)
        vec2 = vec_phi * np.cos(pmesh) + vec_r * np.sin(pmesh)
        pcoord, rcoord = step.geom.p_walls, step.geom.r_centers
        pmesh, rmesh = np.meshgrid(pcoord, rcoord, indexing='ij')
        xmesh, ymesh = rmesh * np.cos(pmesh), rmesh * np.sin(pmesh)
    if step.geom.cartesian:
        xmesh, ymesh = np.meshgrid(xcoord, ycoord, indexing='ij')
    return xmesh, ymesh, vec1, vec2


def plot_scalar(step, var, field=None, axis=None, **extra):
    """Plot scalar field.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
        var (str): the scalar field name.
        field (:class:`numpy.array`): if not None, it is plotted instead of
            step.fields[var].  This is useful to plot a masked or rescaled
            array.  Note that if conf.scaling.dimensional is True, this
            field will be scaled accordingly.
        axis (:class:`matplotlib.axes.Axes`): the axis objet where the field
            should be plotted.  If set to None, a new figure with one subplot
            is created.
        extra (dict): options that will be passed on to
            :func:`matplotlib.axes.Axes.pcolormesh`.
    Returns:
        fig, axis, surf, cbar
            handles to various :mod:`matplotlib` objects, respectively the
            figure, the axis, the surface returned by
            :func:`~matplotlib.axes.Axes.pcolormesh`, and the colorbar returned
            by :func:`matplotlib.pyplot.colorbar`.
    """
    if var in phyvars.FIELD:
        meta = phyvars.FIELD[var]
    else:
        meta = phyvars.FIELD_EXTRA[var]
        meta = phyvars.Varf(_helpers.baredoc(meta.description), meta.dim)
    if step.geom.threed and step.geom.spherical:
        raise NotAvailableError(
            'plot_scalar not implemented for 3D spherical geometry')

    xmesh, ymesh, fld = get_meshes_fld(step, var,
                                       walls=not conf.field.interpolate)
    if conf.field.interpolate and \
       step.geom.spherical and step.geom.twod_yz and \
       fld.shape[0] == step.geom.nytot:
        # add one point to close spherical annulus
        xmesh = np.concatenate((xmesh, xmesh[:1]), axis=0)
        ymesh = np.concatenate((ymesh, ymesh[:1]), axis=0)
        newline = (fld[:1] + fld[-1:]) / 2
        fld = np.concatenate((fld, newline), axis=0)
    xmin, xmax = xmesh.min(), xmesh.max()
    ymin, ymax = ymesh.min(), ymesh.max()

    if field is not None:
        fld = field
    if conf.field.perturbation:
        fld = fld - np.mean(fld, axis=0)
    if conf.field.shift:
        fld = np.roll(fld, conf.field.shift, axis=0)

    fld, unit = step.sdat.scale(fld, meta.dim)

    if axis is None:
        fig, axis = plt.subplots(ncols=1)
    else:
        fig = axis.get_figure()

    if step.sdat.par['magma_oceans_in']['magma_oceans_mode']:
        rcmb = step.sdat.par['geometry']['r_cmb']
        xmax = rcmb + 1
        ymax = xmax
        xmin = -xmax
        ymin = -ymax
        rsurf = xmax if step.timeinfo['thick_tmo'] > 0 \
            else step.geom.r_walls[0, 0, -3]
        cmb = mpat.Circle((0, 0), rcmb, color='dimgray', zorder=0)
        psurf = mpat.Circle((0, 0), rsurf, color='indianred', zorder=0)
        axis.add_patch(psurf)
        axis.add_patch(cmb)

    extra_opts = dict(
        cmap=conf.field.cmap.get(var),
        vmin=conf.plot.vmin,
        vmax=conf.plot.vmax,
        norm=mpl.colors.LogNorm() if var == 'eta' else None,
        rasterized=conf.plot.raster,
        shading='gouraud' if conf.field.interpolate else 'flat',
    )
    extra_opts.update(extra)
    surf = axis.pcolormesh(xmesh, ymesh, fld, **extra_opts)

    cbar = None
    if conf.field.colorbar:
        cax = make_axes_locatable(axis).append_axes(
            'right', size="3%", pad=0.15)
        cbar = plt.colorbar(surf, cax=cax)
        cbar.set_label(meta.description +
                       (' pert.' if conf.field.perturbation else '') +
                       (f' ({unit})' if unit else ''))
    if step.geom.spherical or conf.plot.ratio is None:
        axis.set_aspect('equal')
        axis.set_axis_off()
    else:
        axis.set_aspect(conf.plot.ratio / axis.get_data_ratio())

    axis.set_adjustable('box')
    axis.set_xlim(xmin, xmax)
    axis.set_ylim(ymin, ymax)
    return fig, axis, surf, cbar


def plot_iso(axis, step, var, **extra):
    """Plot isocontours of scalar field.

    Args:
        axis (:class:`matplotlib.axes.Axes`): the axis handler of an
            existing matplotlib figure where the isocontours should
            be plotted.
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
        var (str): the scalar field name.
        extra (dict): options that will be passed on to
            :func:`matplotlib.axes.Axes.contour`.
    """
    xmesh, ymesh, fld = get_meshes_fld(step, var)
    if conf.field.shift:
        fld = np.roll(fld, conf.field.shift, axis=0)
    extra_opts = dict(linewidths=1)
    if 'cmap' not in extra and conf.field.isocolors:
        extra_opts['colors'] = conf.field.isocolors.split(',')
    elif 'colors' not in extra:
        extra_opts['cmap'] = conf.field.cmap.get(var)
    extra_opts.update(extra)
    axis.contour(xmesh, ymesh, fld, **extra_opts)


def plot_vec(axis, step, var):
    """Plot vector field.

    Args:
        axis (:class:`matplotlib.axes.Axes`): the axis handler of an
            existing matplotlib figure where the vector field should
            be plotted.
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
        var (str): the vector field name.
    """
    xmesh, ymesh, vec1, vec2 = get_meshes_vec(step, var)
    dipz = step.geom.nztot // 10
    if conf.field.shift:
        vec1 = np.roll(vec1, conf.field.shift, axis=0)
        vec2 = np.roll(vec2, conf.field.shift, axis=0)
    if step.geom.spherical or conf.plot.ratio is None:
        dipx = dipz
    else:
        dipx = step.geom.nytot if step.geom.twod_yz else step.geom.nxtot
        dipx = int(dipx // 10 * conf.plot.ratio) + 1
    axis.quiver(xmesh[::dipx, ::dipz], ymesh[::dipx, ::dipz],
                vec1[::dipx, ::dipz], vec2[::dipx, ::dipz],
                linewidths=1)


def _findminmax(sdat, sovs):
    """Find min and max values of several fields."""
    minmax = {}
    for step in sdat.walk.filter(snap=True):
        for var in sovs:
            if var in step.fields:
                if var in phyvars.FIELD:
                    dim = phyvars.FIELD[var].dim
                else:
                    dim = phyvars.FIELD_EXTRA[var].dim
                field, _ = sdat.scale(step.fields[var], dim)
                if var in minmax:
                    minmax[var] = (min(minmax[var][0], np.nanmin(field)),
                                   max(minmax[var][1], np.nanmax(field)))
                else:
                    minmax[var] = np.nanmin(field), np.nanmax(field)
    return minmax


def cmd():
    """Implementation of field subcommand.

    Other Parameters:
        conf.field
        conf.core
    """
    sdat = StagyyData()
    lovs = _helpers.list_of_vars(conf.field.plot)
    # no more than two fields in a subplot
    lovs = [[slov[:2] for slov in plov] for plov in lovs]
    minmax = {}
    if conf.plot.cminmax:
        conf.plot.vmin = None
        conf.plot.vmax = None
        sovs = set(slov[0] for plov in lovs for slov in plov)
        minmax = _findminmax(sdat, sovs)
    for step in sdat.walk.filter(snap=True):
        for vfig in lovs:
            fig, axes = plt.subplots(ncols=len(vfig), squeeze=False,
                                     figsize=(6 * len(vfig), 6))
            for axis, var in zip(axes[0], vfig):
                if var[0] not in step.fields:
                    print(f"{var[0]!r} field on snap {step.isnap} not found")
                    continue
                opts = {}
                if var[0] in minmax:
                    opts = dict(vmin=minmax[var[0]][0], vmax=minmax[var[0]][1])
                plot_scalar(step, var[0], axis=axis, **opts)
                if len(var) == 2:
                    if valid_field_var(var[1]):
                        plot_iso(axis, step, var[1])
                    elif valid_field_var(var[1] + '1'):
                        plot_vec(axis, step, var[1])
            if conf.field.timelabel:
                time, unit = sdat.scale(step.timeinfo['t'], 's')
                time = _helpers.scilabel(time)
                axes[0, 0].text(0.02, 1.02, f'$t={time}$ {unit}',
                                transform=axes[0, 0].transAxes)
            oname = '_'.join(chain.from_iterable(vfig))
            plt.tight_layout(w_pad=3)
            _helpers.saveplot(fig, oname, step.isnap)



"""Various helper functions and classes."""

from inspect import getdoc

import matplotlib.pyplot as plt

from . import conf


def out_name(stem, timestep=None):
    """Return StagPy out file name.

    Args:
        stem (str): short description of file content.
        timestep (int): timestep if relevant.

    Returns:
        str: the output file name.

    Other Parameters:
        conf.core.outname (str): the generic name stem, defaults to
            ``'stagpy'``.
    """
    if conf.core.shortname:
        return conf.core.outname
    if timestep is not None:
        stem = f'{stem}{timestep:05d}'
    return conf.core.outname + '_' + stem


def scilabel(value, precision=2):
    """Build scientific notation of some value.

    This is dedicated to use in labels displaying scientific values.

    Args:
        value (float): numeric value to format.
        precision (int): number of decimal digits.

    Returns:
        str: the scientific notation the specified value.
    """
    man, exp = f'{value:.{precision}e}'.split('e')
    exp = int(exp)
    return fr'{man}\times 10^{{{exp}}}'


def saveplot(fig, *name_args, close=True, **name_kwargs):
    """Save matplotlib figure.

    You need to provide :data:`stem` as a positional or keyword argument (see
    :func:`out_name`).

    Args:
        fig (:class:`matplotlib.figure.Figure`): matplotlib figure.
        close (bool): whether to close the figure.
        name_args: positional arguments passed on to :func:`out_name`.
        name_kwargs: keyword arguments passed on to :func:`out_name`.
    """
    oname = out_name(*name_args, **name_kwargs)
    fig.savefig(f'{oname}.{conf.plot.format}',
                format=conf.plot.format, bbox_inches='tight')
    if close:
        plt.close(fig)


def baredoc(obj):
    """Return the first line of the docstring of an object.

    Trailing periods and spaces as well as leading spaces are removed from the
    output.

    Args:
        obj: any Python object.
    Returns:
        str: the first line of the docstring of obj.
    """
    doc = getdoc(obj)
    if not doc:
        return ''
    doc = doc.splitlines()[0]
    return doc.rstrip(' .').lstrip()


def list_of_vars(arg_plot):
    """Construct list of variables per plot.

    Args:
        arg_plot (str): string with variable names separated with
            ``-`` (figures), ``.`` (subplots) and ``,`` (same subplot).
    Returns:
        three nested lists of str

        - variables on the same subplot;
        - subplots on the same figure;
        - figures.
    """
    lovs = [[[var for var in svars.split(',') if var]
             for svars in pvars.split('.') if svars]
            for pvars in arg_plot.split('-') if pvars]
    lovs = [[slov for slov in lov if slov] for lov in lovs if lov]
    return [lov for lov in lovs if lov]


def find_in_sorted_arr(value, array, after=False):
    """Return position of element in a sorted array.

    Returns:
        int: the maximum position i such as array[i] <= value.  If after is
            True, it returns the min i such as value <= array[i] (or 0 if such
            an indices does not exist).
    """
    ielt = array.searchsorted(value)
    if ielt == array.size:
        ielt -= 1
    if not after and array[ielt] != value and ielt > 0:
        ielt -= 1
    return ielt


class CachedReadOnlyProperty:
    """Descriptor implementation of read-only cached properties.

    Properties are cached as ``_cropped_{name}`` instance attribute.

    This is preferable to using a combination of ``@property`` and
    ``@functools.lru_cache`` since the cache is bound to instances and
    therefore get GCd with the instance when the latter is no longer in use
    instead of staying in the cache which would use the instance itself as its
    key.

    This also has an advantage over ``@cached_property`` (Python>3.8): the
    property is read-only instead of being writeable.
    """

    def __init__(self, thunk):
        self._thunk = thunk
        self._name = thunk.__name__
        self._cache_name = f'_cropped_{self._name}'
        self.__doc__ = thunk.__doc__

    def __get__(self, instance, _):
        try:
            return getattr(instance, self._cache_name)
        except AttributeError:
            pass
        cached_value = self._thunk(instance)
        setattr(instance, self._cache_name, cached_value)
        return cached_value

    def __set__(self, instance, _):
        raise AttributeError(
            f'Cannot set {self._name} property of {instance!r}')



import pytest
import stagpy.error
import stagpy.rprof
import stagpy.phyvars


def test_no_rprof_data(sdat):
    with pytest.raises(stagpy.error.MissingDataError):
        sdat.steps[1].rprofs['Tmean']


def test_invalid_rprof(step):
    with pytest.raises(stagpy.error.UnknownRprofVarError):
        step.rprofs['DummyVar']


def test_rprof_bounds_if_no_rprofs(sdat):
    rcmb, rtot = sdat.steps[1].rprofs.bounds
    assert rtot > rcmb


def test_get_rprof(step):
    prof, rad, meta = step.rprofs['Tmean']
    assert rad is step.rprofs.centers
    assert prof.shape == (step.geom.nztot,)
    assert meta == stagpy.phyvars.RPROF['Tmean']


def test_get_rprof_extra(step):
    prof, rad, meta = step.rprofs['diff']
    assert rad is step.rprofs.walls
    assert prof.shape == rad.shape
    assert isinstance(meta, stagpy.phyvars.Varr)



from itertools import chain
from stagpy import phyvars


def test_dim():
    allvars = chain(phyvars.FIELD.values(), phyvars.FIELD_EXTRA.values(),
                    phyvars.RPROF.values(), phyvars.RPROF_EXTRA.values(),
                    phyvars.TIME.values(), phyvars.TIME_EXTRA.values())
    for var in allvars:
        if var.dim != '1':
            assert var.dim in phyvars.SCALES



import pytest
import stagpy.error
import stagpy.field
import stagpy.phyvars


def test_field_unknow(step):
    with pytest.raises(stagpy.error.UnknownFieldVarError):
        step.fields['InvalidField']
    assert 'InvalidField' not in step.fields


def test_field_missing(step):
    with pytest.raises(stagpy.error.MissingDataError):
        step.fields['rsc']
    assert 'rsc' not in step.fields


def test_valid_field_var():
    for var in stagpy.phyvars.FIELD:
        assert stagpy.field.valid_field_var(var)
    for var in stagpy.phyvars.FIELD_EXTRA:
        assert stagpy.field.valid_field_var(var)


def test_valid_field_var_invalid():
    assert not stagpy.field.valid_field_var('dummyfieldvar')


def test_get_meshes_fld(step):
    xmesh, ymesh, fld = stagpy.field.get_meshes_fld(step, 'T')
    assert len(fld.shape) == 2
    assert xmesh.shape == ymesh.shape == fld.shape


def test_get_meshes_vec(step):
    xmesh, ymesh, vec1, vec2 = stagpy.field.get_meshes_vec(step, 'v')
    assert len(vec1.shape) == 2
    assert xmesh.shape == ymesh.shape == vec1.shape == vec2.shape



"""Define high level structure StagyyData and helper classes.

Note:
    The helper classes are not designed to be instantiated on their own, but
    only as attributes of StagyyData instances. Users of this module should
    only instantiate :class:`StagyyData`.

"""

import re
import pathlib
from collections import namedtuple
from itertools import zip_longest

import numpy as np

from . import conf, error, parfile, phyvars, stagyyparsers, _helpers, _step
from ._helpers import CachedReadOnlyProperty as crop


def _as_view_item(obj):
    """Return None or a suitable iterable to build a _StepsView."""
    try:
        iter(obj)
        return obj
    except TypeError:
        pass
    if isinstance(obj, slice):
        return (obj,)


class _Scales:
    """Dimensionful scales.

    Args:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Scales` instance.
    """

    def __init__(self, sdat):
        self._sdat = sdat

    @property
    def length(self):
        """Length in m."""
        return self._sdat.par['geometry']['d_dimensional']

    @property
    def temperature(self):
        """Temperature in K."""
        return self._sdat.par['refstate']['deltaT_dimensional']

    @property
    def density(self):
        """Density in kg/m3."""
        return self._sdat.par['refstate']['dens_dimensional']

    @property
    def th_cond(self):
        """Thermal conductivity in W/(m.K)."""
        return self._sdat.par['refstate']['tcond_dimensional']

    @property
    def sp_heat(self):
        """Specific heat capacity in J/(kg.K)."""
        return self._sdat.par['refstate']['Cp_dimensional']

    @property
    def dyn_visc(self):
        """Dynamic viscosity in Pa.s."""
        return self._sdat.par['viscosity']['eta0']

    @property
    def th_diff(self):
        """Thermal diffusivity in m2/s."""
        return self.th_cond / (self.density * self.sp_heat)

    @property
    def time(self):
        """Time in s."""
        return self.length**2 / self.th_diff

    @property
    def power(self):
        """Power in W."""
        return self.th_cond * self.temperature * self.length

    @property
    def heat_flux(self):
        """Local heat flux in W/m2."""
        return self.power / self.length**2

    @property
    def stress(self):
        """Stress in Pa."""
        return self.dyn_visc / self.time


class _Refstate:
    """Reference state profiles.

    The :attr:`StagyyData.refstate` attribute is an instance of this class.
    Reference state profiles are accessed through the attributes of this
    object.

    Args:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Steps` instance.
    """

    def __init__(self, sdat):
        self._sdat = sdat

    @crop
    def _data(self):
        """Read reference state profile."""
        reffile = self._sdat.filename('refstat.dat')
        if self._sdat.hdf5 and not reffile.is_file():
            # check legacy folder as well
            reffile = self._sdat.filename('refstat.dat', force_legacy=True)
        return stagyyparsers.refstate(reffile)

    @property
    def systems(self):
        """Reference state profiles of phases.

        It is a list of list of :class:`pandas.DataFrame` containing
        the reference profiles associated with each phase in each system.

        Example:
            The temperature profile of the 3rd phase in the 1st
            system is

            >>> sdat.refstate.systems[0][2]['T']
        """
        return self._data[0]

    @property
    def adiabats(self):
        """Adiabatic reference state profiles.

        It is a list of :class:`pandas.DataFrame` containing the reference
        profiles associated with each system.  The last item is the combined
        adiabat.

        Example:
            The adiabatic temperature profile of the 2nd system is

            >>> sdat.refstate.adiabats[1]['T']

            The combined density profile is

            >>> sdat.refstate.adiabats[-1]['rho']
        """
        return self._data[1]


Tseries = namedtuple('Tseries', ['values', 'time', 'meta'])


class _Tseries:
    """Time series.

    The :attr:`StagyyData.tseries` attribute is an instance of this class.

    :class:`_Tseries` implements the getitem mechanism.  Keys are series names
    defined in :data:`stagpy.phyvars.TIME[_EXTRA]`.  An item is a named tuple
    ('values', 'time', 'meta'), respectively the series itself, the time at
    which it is evaluated, and meta is a :class:`stagpy.phyvars.Vart` instance
    with relevant metadata.  Note that series are automatically scaled if
    conf.scaling.dimensional is True.

    Attributes:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Tseries` instance.
    """

    def __init__(self, sdat):
        self.sdat = sdat
        self._cached_extra = {}

    @crop
    def _data(self):
        timefile = self.sdat.filename('TimeSeries.h5')
        data = stagyyparsers.time_series_h5(
            timefile, list(phyvars.TIME.keys()))
        if data is not None:
            return data
        timefile = self.sdat.filename('time.dat')
        if self.sdat.hdf5 and not timefile.is_file():
            # check legacy folder as well
            timefile = self.sdat.filename('time.dat', force_legacy=True)
        data = stagyyparsers.time_series(timefile, list(phyvars.TIME.keys()))
        return data

    @property
    def _tseries(self):
        if self._data is None:
            raise error.MissingDataError(f'No tseries data in {self.sdat}')
        return self._data

    def __getitem__(self, name):
        if name in self._tseries.columns:
            series = self._tseries[name].values
            time = self.time
            if name in phyvars.TIME:
                meta = phyvars.TIME[name]
            else:
                meta = phyvars.Vart(name, None, '1')
        elif name in self._cached_extra:
            series, time, meta = self._cached_extra[name]
        elif name in phyvars.TIME_EXTRA:
            meta = phyvars.TIME_EXTRA[name]
            series, time = meta.description(self.sdat)
            meta = phyvars.Vart(_helpers.baredoc(meta.description),
                                meta.kind, meta.dim)
            self._cached_extra[name] = series, time, meta
        else:
            raise error.UnknownTimeVarError(name)
        series, _ = self.sdat.scale(series, meta.dim)
        time, _ = self.sdat.scale(time, 's')
        return Tseries(series, time, meta)

    def tslice(self, name, tstart=None, tend=None):
        """Return a Tseries between specified times.

        Args:
            name (str): time variable.
            tstart (float): starting time. Set to None to start at the
                beginning of available data.
            tend (float): ending time. Set to None to stop at the end of
                available data.
        """
        data, time, meta = self[name]
        istart = 0
        iend = len(time)
        if tstart is not None:
            istart = _helpers.find_in_sorted_arr(tstart, time)
        if tend is not None:
            iend = _helpers.find_in_sorted_arr(tend, time, True) + 1
        return Tseries(data[istart:iend], time[istart:iend], meta)

    @property
    def time(self):
        """Time vector."""
        return self._tseries['t'].values

    @property
    def isteps(self):
        """Step indices.

        This is such that time[istep] is at step isteps[istep].
        """
        return self._tseries.index.values

    def at_step(self, istep):
        """Time series output for a given step."""
        return self._tseries.loc[istep]


class _RprofsAveraged(_step._Rprofs):
    """Radial profiles time-averaged over a :class:`_StepsView`.

    The :attr:`_StepsView.rprofs_averaged` attribute is an instance of this
    class.

    It implements the same interface as :class:`~stagpy._step._Rprofs` but
    returns time-averaged profiles instead.

    Attributes:
        steps (:class:`_StepsView`): the object owning the
            :class:`_RprofsAveraged` instance
    """

    def __init__(self, steps):
        self.steps = steps.filter(rprofs=True)
        self._cached_data = {}
        super().__init__(next(iter(self.steps)))

    def __getitem__(self, name):
        # the averaging method has two shortcomings:
        # - does not take into account time changing geometry;
        # - does not take into account time changing timestep.
        if name in self._cached_data:
            return self._cached_data[name]
        steps_iter = iter(self.steps)
        rprof, rad, meta = next(steps_iter).rprofs[name]
        rprof = np.copy(rprof)
        nprofs = 1
        for step in steps_iter:
            nprofs += 1
            rprof += step.rprofs[name].values
        rprof /= nprofs
        self._cached_data[name] = _step.Rprof(rprof, rad, meta)
        return self._cached_data[name]


class _Steps:
    """Collections of time steps.

    The :attr:`StagyyData.steps` attribute is an instance of this class.
    Time steps (which are :class:`~stagpy._step.Step` instances) can be
    accessed with the item accessor::

        sdat = StagyyData('path/to/run')
        sdat.steps[istep]  # Step object of the istep-th time step

    Slices or tuple of istep and slices of :class:`_Steps` object are
    :class:`_StepsView` instances that you can iterate and filter::

        for step in steps[500:]:
            # iterate through all time steps from the 500-th one
            do_something(step)

        for step in steps[-100:].filter(snap=True):
            # iterate through all snapshots present in the last 100 time steps
            do_something(step)

        for step in steps[0,3,5,-2:]:
            # iterate through steps 0, 3, 5 and the last two
            do_something(step)

    Args:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Steps` instance.
    Attributes:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Steps` instance.
    """

    def __init__(self, sdat):
        self.sdat = sdat
        self._data = {}

    def __repr__(self):
        return f'{self.sdat!r}.steps'

    def __getitem__(self, istep):
        keys = _as_view_item(istep)
        if keys is not None:
            return _StepsView(self, keys)
        try:
            istep = int(istep)
        except ValueError:
            raise error.InvalidTimestepError(
                self.sdat, istep, 'Time step should be an integer value')
        if istep < 0:
            istep += len(self)
            if istep < 0:
                istep -= len(self)
                raise error.InvalidTimestepError(
                    self.sdat, istep,
                    f'Last istep is {len(self) - 1}')
        if istep not in self._data:
            self._data[istep] = _step.Step(istep, self.sdat)
        return self._data[istep]

    def __delitem__(self, istep):
        if istep is not None and istep in self._data:
            self.sdat._collected_fields = [
                (i, f) for i, f in self.sdat._collected_fields if i != istep]
            del self._data[istep]

    @crop
    def _len(self):
        # not necessarily the last one...
        return self.sdat.tseries.isteps[-1] + 1

    def __len__(self):
        return self._len

    def __iter__(self):
        return iter(self[:])

    def at_time(self, time, after=False):
        """Return step corresponding to a given physical time.

        Args:
            time (float): the physical time requested.
            after (bool): when False (the default), the returned step is such
                that its time is immediately before the requested physical
                time. When True, the returned step is the next one instead (if
                it exists, otherwise the same step is returned).

        Returns:
            :class:`~stagpy._step.Step`: the relevant step.
        """
        itime = _helpers.find_in_sorted_arr(time, self.sdat.tseries.time,
                                            after)
        return self[self.sdat.tseries.isteps[itime]]

    def filter(self, **filters):
        """Build a _StepsView with requested filters."""
        return self[:].filter(**filters)


class _Snaps(_Steps):
    """Collections of snapshots.

    The :attr:`StagyyData.snaps` attribute is an instance of this class.
    Snapshots (which are :class:`~stagpy._step.Step` instances) can be accessed
    with the item accessor::

        sdat = StagyyData('path/to/run')
        sdat.snaps[isnap]  # Step object of the isnap-th snapshot

    This class inherits from :class:`_Steps`.

    Args:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Snaps` instance.
    Attributes:
        sdat (:class:`StagyyData`): the StagyyData instance owning the
            :class:`_Snaps` instance.
    """

    def __init__(self, sdat):
        self._isteps = {}
        self._all_isteps_known = False
        super().__init__(sdat)

    def __repr__(self):
        return f'{self.sdat!r}.snaps'

    def __getitem__(self, isnap):
        keys = _as_view_item(isnap)
        if keys is not None:
            return _StepsView(self, keys).filter(snap=True)
        if isnap < 0:
            isnap += len(self)
        if isnap < 0 or isnap >= len(self):
            istep = None
        else:
            istep = self._isteps.get(
                isnap, None if self._all_isteps_known else -1)
        if istep == -1:
            # isnap not in _isteps but not all isteps known, keep looking
            binfiles = self.sdat._binfiles_set(isnap)
            if binfiles:
                istep = stagyyparsers.fields(binfiles.pop(), only_istep=True)
            else:
                istep = None
            if istep is not None:
                self._bind(isnap, istep)
            else:
                self._isteps[isnap] = None
        if istep is None:
            raise error.InvalidSnapshotError(
                self.sdat, isnap, 'Invalid snapshot index')
        return self.sdat.steps[istep]

    def __delitem__(self, isnap):
        istep = self._isteps.get(isnap)
        del self.sdat.steps[istep]

    @crop
    def _len(self):
        length = -1
        if self.sdat.hdf5:
            isnap = -1
            for isnap, istep in stagyyparsers.read_time_h5(self.sdat.hdf5):
                self._bind(isnap, istep)
            length = isnap
            self._all_isteps_known = True
        if length < 0:
            out_stem = re.escape(pathlib.Path(
                self.sdat.par['ioin']['output_file_stem'] + '_').name[:-1])
            rgx = re.compile(f'^{out_stem}_([a-zA-Z]+)([0-9]{{5}})$')
            fstems = set(fstem for fstem in phyvars.FIELD_FILES)
            for fname in self.sdat._files:
                match = rgx.match(fname.name)
                if match is not None and match.group(1) in fstems:
                    length = max(int(match.group(2)), length)
        if length < 0:
            raise error.NoSnapshotError(self.sdat)
        return length + 1

    def at_time(self, time, after=False):
        """Return snap corresponding to a given physical time.

        Args:
            time (float): the physical time requested.
            after (bool): when False (the default), the returned snap is such
                that its time is immediately before the requested physical
                time. When True, the returned snap is the next one instead (if
                it exists, otherwise the same snap is returned).

        Returns:
            :class:`~stagpy._step.Step`: the relevant snap.
        """
        # in theory, this could be a valid implementation of _Steps.at_time
        # but this isn't safe against missing data...
        igm = 0
        igp = len(self) - 1
        while igp - igm > 1:
            istart = igm + (igp - igm) // 2
            if self[istart].time >= time:
                igp = istart
            else:
                igm = istart
        if self[igp].time > time and not after and igp > 0:
            igp -= 1
        return self[igp]

    def _bind(self, isnap, istep):
        """Register the isnap / istep correspondence.

        Args:
            isnap (int): snapshot index.
            istep (int): time step index.
        """
        self._isteps[isnap] = istep
        self.sdat.steps[istep]._isnap = isnap


class _StepsView:
    """Filtered iterator over steps or snaps.

    Instances of this class are returned when taking slices of
    :attr:`StagyyData.steps` or :attr:`StagyyData.snaps` attributes.

    Args:
        steps_col (:class:`_Steps` or :class:`_Snaps`): steps collection,
            i.e. :attr:`StagyyData.steps` or :attr:`StagyyData.snaps`
            attributes.
        items (iterable): iterable of isteps/isnaps or slices.
    """

    def __init__(self, steps_col, items):
        self._col = steps_col
        self._items = items
        self._rprofs_averaged = None
        self._flt = {
            'snap': False,
            'rprofs': False,
            'fields': [],
            'func': lambda _: True,
        }
        self._dflt_func = self._flt['func']

    @property
    def rprofs_averaged(self):
        """Time-averaged radial profiles."""
        if self._rprofs_averaged is None:
            self._rprofs_averaged = _RprofsAveraged(self)
        return self._rprofs_averaged

    @crop
    def stepstr(self):
        """String representation of the requested set of steps."""
        items = []
        no_slice = True
        for item in self._items:
            if isinstance(item, slice):
                items.append('{}:{}:{}'.format(*item.indices(len(self._col))))
                no_slice = False
            else:
                items.append(repr(item))
        item_str = ','.join(items)
        if no_slice and len(items) == 1:
            item_str += ','
        colstr = repr(self._col).rsplit('.', maxsplit=1)[-1]
        return f'{colstr}[{item_str}]'

    def __repr__(self):
        rep = f'{self._col.sdat!r}.{self.stepstr}'
        flts = []
        for flt in ('snap', 'rprofs', 'fields'):
            if self._flt[flt]:
                flts.append(f'{flt}={self._flt[flt]!r}')
        if self._flt['func'] is not self._dflt_func:
            flts.append(f"func={self._flt['func']!r}")
        if flts:
            rep += '.filter({})'.format(', '.join(flts))
        return rep

    def _pass(self, item):
        """Check whether an item passes the filters."""
        try:
            step = self._col[item]
        except KeyError:
            return False
        okf = True
        okf = okf and (not self._flt['snap'] or step.isnap is not None)
        if self._flt['rprofs']:
            try:
                _ = step.rprofs.centers
            except error.MissingDataError:
                return False
        okf = okf and all(f in step.fields for f in self._flt['fields'])
        okf = okf and bool(self._flt['func'](step))
        return okf

    def filter(self, **filters):
        """Update filters with provided arguments.

        Note that filters are only resolved when the view is iterated, and
        hence they do not compose. Each call to filter merely updates the
        relevant filters. For example, with this code::

            view = sdat.steps[500:].filter(rprofs=True, fields=['T'])
            view.filter(fields=[])

        the produced ``view``, when iterated, will generate the steps after the
        500-th that have radial profiles. The ``fields`` filter set in the
        first line is emptied in the second line.

        Args:
            snap (bool): the step must be a snapshot to pass.
            rprofs (bool): the step must have rprofs data to pass.
            fields (list): list of fields that must be present to pass.
            func (function): arbitrary function taking a
                :class:`~stagpy._step.Step` as argument and returning a True
                value if the step should pass the filter.

        Returns:
            self.
        """
        for flt, val in self._flt.items():
            self._flt[flt] = filters.pop(flt, val)
        if filters:
            raise error.UnknownFiltersError(filters.keys())
        return self

    def __iter__(self):
        for item in self._items:
            if isinstance(item, slice):
                idx = item.indices(len(self._col))
                yield from (self._col[i] for i in range(*idx)
                            if self._pass(i))
            elif self._pass(item):
                yield self._col[item]

    def __eq__(self, other):
        return all(s1 is s2 for s1, s2 in zip_longest(self, other))


class StagyyData:
    """Generic lazy interface to StagYY output data.

    Args:
        path (pathlike): path of the StagYY run. It can either be the path
            of the directory containing the par file, or the path of the
            par file. If the path given is a directory, the path of the par
            file is assumed to be path/par.  If no path is given (or None)
            it is set to ``conf.core.path``.

    Other Parameters:
        conf.core.path: the default path.

    Attributes:
        steps (:class:`_Steps`): collection of time steps.
        snaps (:class:`_Snaps`): collection of snapshots.
        scales (:class:`_Scales`): dimensionful scaling factors.
        refstate (:class:`_Refstate`): reference state profiles.
    """

    def __init__(self, path=None):
        if path is None:
            path = conf.core.path
        runpath = pathlib.Path(path)
        if runpath.is_file():
            parname = runpath.name
            runpath = runpath.parent
        else:
            parname = 'par'
        self._rundir = {'path': runpath,
                        'par': parname}
        self._par = parfile.readpar(self.parpath, self.path)
        self.scales = _Scales(self)
        self.refstate = _Refstate(self)
        self.tseries = _Tseries(self)
        self.steps = _Steps(self)
        self.snaps = _Snaps(self)
        self._nfields_max = 50
        # list of (istep, field_name) in memory
        self._collected_fields = []

    def __repr__(self):
        return f'StagyyData({self.path!r})'

    def __str__(self):
        return f'StagyyData in {self.path}'

    @property
    def path(self):
        """Path of StagYY run directory.

        :class:`pathlib.Path` instance.
        """
        return self._rundir['path']

    @property
    def parpath(self):
        """Path of par file.

        :class:`pathlib.Path` instance.
        """
        return self.path / self._rundir['par']

    @crop
    def hdf5(self):
        """Path of output hdf5 folder if relevant, None otherwise."""
        h5_folder = self.path / self.par['ioin']['hdf5_output_folder']
        if (h5_folder / 'Data.xmf').is_file():
            return h5_folder

    @property
    def par(self):
        """Content of par file.

        This is a :class:`f90nml.namelist.Namelist`, the first key being
        namelists and the second key the parameter name.
        """
        return self._par

    @crop
    def _rprof_and_times(self):
        rproffile = self.filename('rprof.h5')
        data = stagyyparsers.rprof_h5(rproffile, list(phyvars.RPROF.keys()))
        if data[1] is not None:
            return data
        rproffile = self.filename('rprof.dat')
        if self.hdf5 and not rproffile.is_file():
            # check legacy folder as well
            rproffile = self.filename('rprof.dat', force_legacy=True)
        return stagyyparsers.rprof(rproffile, list(phyvars.RPROF.keys()))

    @property
    def rtimes(self):
        """Radial profiles times.

        This is a :class:`pandas.DataFrame` with istep as index.
        """
        return self._rprof_and_times[1]

    @crop
    def _files(self):
        """Set of found binary files output by StagYY."""
        out_stem = pathlib.Path(self.par['ioin']['output_file_stem'] + '_')
        out_dir = self.path / out_stem.parent
        if out_dir.is_dir():
            return set(out_dir.iterdir())
        return set()

    @property
    def walk(self):
        """Return view on configured steps slice.

        Other Parameters:
            conf.core.snapshots: the slice of snapshots.
            conf.core.timesteps: the slice of timesteps.
        """
        if conf.core.snapshots:
            return self.snaps[conf.core.snapshots]
        elif conf.core.timesteps:
            return self.steps[conf.core.timesteps]
        return self.snaps[-1, ]

    @property
    def nfields_max(self):
        """Maximum number of scalar fields kept in memory.

        Setting this to a value lower or equal to 5 raises a
        :class:`~stagpy.error.InvalidNfieldsError`.  Set this to ``None`` if
        you do not want any limit on the number of scalar fields kept in
        memory.  Defaults to 50.
        """
        return self._nfields_max

    @nfields_max.setter
    def nfields_max(self, nfields):
        """Check nfields > 5 or None."""
        if nfields is not None and nfields <= 5:
            raise error.InvalidNfieldsError(nfields)
        self._nfields_max = nfields

    def scale(self, data, unit):
        """Scales quantity to obtain dimensionful quantity.

        Args:
            data (numpy.array): the quantity that should be scaled.
            dim (str): the dimension of data as defined in phyvars.
        Return:
            (float, str): scaling factor and unit string.
        Other Parameters:
            conf.scaling.dimensional: if set to False (default), the factor is
                always 1.
        """
        if self.par['switches']['dimensional_units'] or \
           not conf.scaling.dimensional or \
           unit == '1':
            return data, ''
        scaling = phyvars.SCALES[unit](self.scales)
        factor = conf.scaling.factors.get(unit, ' ')
        if conf.scaling.time_in_y and unit == 's':
            scaling /= conf.scaling.yearins
            unit = 'yr'
        elif conf.scaling.vel_in_cmpy and unit == 'm/s':
            scaling *= 100 * conf.scaling.yearins
            unit = 'cm/y'
        if factor in phyvars.PREFIXES:
            scaling *= 10**(-3 * (phyvars.PREFIXES.index(factor) + 1))
            unit = factor + unit
        return data * scaling, unit

    def filename(self, fname, timestep=None, suffix='', force_legacy=False):
        """Return name of StagYY output file.

        Args:
            fname (str): name stem.
            timestep (int): snapshot number, set to None if this is not
                relevant.
            suffix (str): optional suffix of file name.
            force_legacy (bool): force returning the legacy output path.
        Returns:
            :class:`pathlib.Path`: the path of the output file constructed
            with the provided segments.
        """
        if timestep is not None:
            fname += f'{timestep:05d}'
        fname += suffix
        if not force_legacy and self.hdf5:
            fpath = self.hdf5 / fname
        else:
            fpath = self.par['ioin']['output_file_stem'] + '_' + fname
            fpath = self.path / fpath
        return fpath

    def _binfiles_set(self, isnap):
        """Set of existing binary files at a given snap.

        Args:
            isnap (int): snapshot index.
        Returns:
            set of pathlib.Path: the set of output files available for this
            snapshot number.
        """
        possible_files = set(self.filename(fstem, isnap, force_legacy=True)
                             for fstem in phyvars.FIELD_FILES)
        return possible_files & self._files



import re
import pytest
import subprocess


@pytest.fixture(params=[('ra-100000', 5, 1000), ('annulus', 100, 3000)])
def dir_isnap(request, repo_dir):
    return (repo_dir / 'Examples' / request.param[0],
            request.param[1], request.param[2])


@pytest.fixture(params=[
    ('stagpy field', ['stagpy_T_stream{:05d}.pdf']),
    ('stagpy field -o=T.v3', ['stagpy_T_v3{:05d}.pdf']),
    ('stagpy field -o=T-v3', ['stagpy_T{:05d}.pdf',
                              'stagpy_v3{:05d}.pdf']),
])
def all_cmd_field(request, dir_isnap):
    cmd = request.param[0]
    cmd += ' -p={}'.format(dir_isnap[0])
    expected_files = []
    for expfile in request.param[1]:
        expected_files.append(expfile.format(dir_isnap[1]))
    return cmd, expected_files


@pytest.fixture(params=[
    ('stagpy rprof', ['stagpy_rprof_Tmean_{}.pdf']),
    ('stagpy rprof -o=Tmean,vzabs', ['stagpy_rprof_Tmean_vzabs_{}.pdf']),
])
def all_cmd_rprof(request, dir_isnap):
    cmd = request.param[0]
    cmd += ' -p={}'.format(dir_isnap[0])
    expected_files = []
    for expfile in request.param[1]:
        expected_files.append(expfile.format(dir_isnap[2]))
    return cmd, expected_files


@pytest.fixture(params=[
    ('stagpy time -o Tmean.Nutop', ['stagpy_time_Tmean_Nutop.pdf']),
])
def all_cmd_time(request, dir_isnap):
    cmd = request.param[0]
    cmd += ' -p={}'.format(dir_isnap[0])
    return cmd, request.param[1]


def helper_test_cli(all_cmd, tmp):
    subprocess.run(all_cmd[0] + ' -n={}/stagpy'.format(tmp), shell=True)
    produced_files = tmp.listdir(sort=True)
    expected_files = [tmp / expfile for expfile in all_cmd[1]]
    assert produced_files == expected_files


def test_field_cli(all_cmd_field, tmpdir):
    helper_test_cli(all_cmd_field, tmpdir)


def test_rprof_cli(all_cmd_rprof, tmpdir):
    helper_test_cli(all_cmd_rprof, tmpdir)


def test_time_cli(all_cmd_time, tmpdir):
    helper_test_cli(all_cmd_time, tmpdir)


def test_err_cli():
    subp = subprocess.run('stagpy field', shell=True, stderr=subprocess.PIPE)
    reg = re.compile(br'^Oops!.*\nPlease.*\n\nNoParFileError.*$')
    assert reg.match(subp.stderr)



import os

from setuptools import setup

with open('README.rst') as rdm:
    README = rdm.read()

DEPENDENCIES = [
    'loam>=0.3.1',
    'f90nml>=1.3.1',
    'setuptools_scm>=6.2',
]
HEAVY = [
    'numpy>=1.19',
    'scipy>=1.5',
    'pandas>=1.1',
    'h5py>=3.1',
    'matplotlib>=3.3',
]

ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'
# heavy dependencies are mocked out on Read the Docs
if not ON_RTD:
    DEPENDENCIES.extend(HEAVY)

setup(
    name='stagpy',

    description='Tool for StagYY output files processing',
    long_description=README,

    url='https://github.com/StagPython/StagPy',

    author='Martina Ulvrova, Adrien Morison, Stéphane Labrosse',
    author_email='adrien.morison@gmail.com',

    license='Apache',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    python_requires='>=3.6',
    packages=['stagpy'],
    entry_points={
        'console_scripts': ['stagpy = stagpy.__main__:main']
    },
    include_package_data=True,
    install_requires=DEPENDENCIES,
)
